! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!************************************************
!
! Program to check the OM background-determination
! routines. Square apertures are placed on the
! image and the total counts within it obtained/
! The background is obtained by using the pixels
! within the aperture and using one of several
! backgroud methods. The background is subtracted
! from the total counts. For an image with no 
! sources, the mean of these background-subtracted
! counts should be zero.
! 
!***********************************************
module ombackground_mod
use dal
use cal
use errorhandling
use omconstants_mod
!use utilities_mod
use omutility_mod
use graph_mod
use omphotometry_mod
use omskybackground_mod
use graph_mod
use getfilterid_mod
use ommode_mod
use robustscale_mod
use backgroundimage_mod
use omnumerics_mod
use statistics_mod
implicit none


real(single), private, pointer :: theimage(:, :) => null()
logical, private :: plotGraph = .false.
integer(int8), private :: nMethods = 0

contains

!****************************************************************
! 
!
!
!
!
!****************************************************************
subroutine ombackground

   
    character(256) :: imageFile, backgroundImageFile
  
   
    type(datasetT) :: dataSetH
    integer(int16) :: nx, ny
  
      
    real(single) :: sigma
    logical :: binBpe  
    character(500) :: title
    character(5) :: filter
    character(256) :: plotFile
 


    nMethods = getNumberBackgroundMethods() ! In mssllib, omskybackground_mod
 
    nMethods = nMethods - 2
    call setBackgroundNsigma( 2.0 )

    imageFile = trim(stringParameter("imageset"))


    dataSetH = dataSet( imageFile, READ)
    filter = stringAttribute( dataSetH, "FILTER")
 

    ! compute binning
    binx = int32Attribute( dataSetH, "BINAX1")
    biny = int32Attribute( dataSetH, "BINAX2")
    binBPE = booleanAttribute( dataSetH, "BINBPE")

    binx = 2 ** binx
  
    if (binBPE) then
        binx = binx * 2
     
    end if


    call GetImage(dataSetH, nx, ny, theimage)

     

   
    call release(dataSetH)

    write(title, '(a, a, a)' ) trim(imageFile), " Filter ", trim(filter)
   


    !********************************************************
    !
    ! Compute the global background (counts per pixel)
    !
    !********************************************************
    backgroundCountsMode = ComputeBackgroundMode2(theimage, nx, ny, 3.0, sigma)





    !*******************************************************
    !
    ! Draw graphs for point-soure detection
    !
    !*******************************************************
    plotFile = trim(stringParameter("plotfile"))
    if( len(plotFile) > 2) then
        call pgbegin(0, trim(stringParameter("plotfile")) // '/CPS', 1, 1)
        call SetUpDefaultcolourIndices( )
        plotGraph = .true.
    else
        plotGraph = .false.
   
    end if 


    call drawGraphs1( theImage, title )  

   
    !******************************************************************
    !
    ! Do the background checks on this image
    !
    !*******************************************************************
    write(errstr, '(a)' ) &
          "Background method comparisons excluding sources"
    call message(AppMsg, Verbose)


    call backgroundCheck(3.0, 0.0, "Background method comparisons excluding sources")
    
    !******************************************************************
    !
    ! Do the background checks on this image
    !
    !*******************************************************************
    write(errstr, '(a)' ) &
          "Background method comparisons including sources"
    call message(AppMsg, Verbose)

    call backgroundCheck(3.0, 1000000.0, "Background method comparisons including sources")
    
    call pgclos()
 




    backgroundImageFile = trim( stringparameter( "backgroundset" ))

    if( len(trim(backgroundImageFile)) > 3) then


    
        !*********************************************************
        ! NB theImage is the actual image on input, on output
        ! it is the backgroudn image
        !
        !********************************************************
        call createTheBackgroundImage( theImage, imageFile, backgroundImageFile )
        
   
    end if




    deallocate(theimage)

end subroutine ombackground





!*************************************************************************
!
!
! THE ACTUAL ROUTINE
!
!
!*************************************************************************
subroutine backgroundCheck(nSigma, maxHeight, title)
    real(single), intent(in) :: nSigma, maxHeight
    character(*), intent(in) :: title  
    integer :: i, j, k, n, n1, nx, ny, il, iu, jl, ju, imin, imax, jmin, jmax, nPix
  
  
    real(single) :: meanValue, std
  
 
    integer :: dx
   
    integer, parameter :: nMax = 100000

    real(double), dimension(10)  ::  sum
    real(single), pointer, dimension(:, :)  :: counts => null(), background => null()
    real(single) :: p1, p2 
    character(100) :: xLabel

     

    nx = ubound(theImage, 1)
    ny = ubound(theImage, 2)

    dx = 10 / binx
    imin = dx + 1
    imax = nx - dx - 1
    jmin = dx + 1
    jmax = ny - dx - 1
  
    nPix = (2 * dx + 1) * (2 * dx + 1)

    n = 1



    background => allocate2dSingle(1, nMax, 1, int(nMethods, int32), "background")
    counts => allocate2dSingle(1, nMax, 1, int(nMethods, int32), "counts")



iLoop :    do i = iMin, iMax, dx
!        write(*, *) "Backgroundcheck1 line 315, i=", i

        il = i - dx
        iu = i + dx

        do j = jMin, jMax, dx
            jl = j - dx
            ju = j + dx
            if( imageOkay(theImage(il : iu, jl : ju), maxHeight) ) then
                !****************************************************
                !
                !
                !****************************************************  
                call computeTheBackground(theImage(il : iu, jl : ju), nsigma, background(n, :), &
                        sum, n1)

                if(n1 > 0) then                                  
                    do k = 1, nMethods       
                  
                        !*****************************************************
                        !
                        ! The mean background-subtracted counts
                        !
                        !
                        !*****************************************************                        
                        counts(n, k) = (sum(k) - dble(n1) * dble(background(n, k))) / dble(n1)
                    
                    end do 
                    if( n .eq. nMax) then
                        exit iLoop
                    end if
                    n = n + 1
                end if  
            end if  
        end do
    end do iLoop



    if(n > 1) then

        write(errstr, '(a, f0.2, a, i0, a, i0, a, i0)' ) &
          "Background method comparisons, global background=", backgroundCountsMode
        call message(AppMsg, Verbose)



        write(errstr, '(a, i0, a, i0, a, i0, a)' ) &
    
          "Number of image apertures=", n, " (aperture width=", (2 * dx + 1), &
          " number of pixels in aperture=", nPix, ")" 

        call message(AppMsg, Verbose)




        do i = 1, nMethods
    
         
            call ComputeAverage( background(:, i), n, meanValue, std)
            p1 = Percentile(n, background(:, i), 0.05)
            p2 = Percentile(n, background(:, i), 0.95)


            !************************************************************
            !
            ! Compute the mean background and standard deviation
            !
            !***********************************************************      

            write(errstr, '(a, a)' ) "Background method = ", trim(getBackgroundText(i))   ! getBackgroundText is in mssllin, omskybackground
            call message(AppMsg, Verbose)




            write(errstr, '(a, f0.2, a, f0.2, a, f0.2, a, f0.2, a, i0, a)' ) "Mean computed background=", meanValue, "+/-", std, &
                  " 0.05 percentile=", p1, " 0.95 percentile=", p2, " (counts per pixel) (n=", n, ")"
            call message(AppMsg, Verbose)



            call ComputeAverage( counts(:, i), n, meanValue, std)


            meanValue = meanValue / dble(nPix)
            stD = stD / dble(nPix)
            p1 = Percentile(n, counts(:, i), 0.05) / dble(nPix)
            p2 = Percentile(n, counts(:, i), 0.95) / dble(nPix)





            write(errstr, '(a, f0.4, a, f0.4, a, f0.4, a, f0.4, a, i0, a)' ) &
                 "Mean background-subtracted counts per pixel=", meanValue, "+/-", std, &
                 " 0.05 percentile=", p1, " 0.95 percentile=", p2, " (n=", n, ")"
        
            call message(AppMsg, Verbose)




        end do
 



        if(plotGraph) then

            xLabel = "Computed background (counts per pixel)"        
            call drawGraphs2( background, n, xLabel, title )

            xLabel = "Background subtracted aperture counts (counts per pixel)"        
    
            call drawGraphs2( counts, n, xLabel, title )

        end if





   else

        write(errstr, '(a)' ) &
          "Crowded field- too few blank regions to determine background"
       call message(AppMsg, Verbose)


 
    end if

 


    deallocate( backGround, counts )
    nullify( backGround, counts )

end subroutine backgroundCheck

!***********************************************************
!
!
!
!
!**********************************************************
logical function imageOkay(image, maxHeight1)
    real(single) :: image(:, :)
    real(single), intent(in), optional :: maxHeight1
    integer :: il, iu, jl, ju, i, j
    real(single) :: maxHeight

    il = lbound(image, 1)
    iu = ubound(image, 1)
    jl = lbound(image, 2)
    ju = ubound(image, 2)

    if(present(maxHeight1) ) then
        if(maxHeight1 > 0.0) then
            maxheight = maxHeight1
        else
            maxHeight = backgroundCountsMode + max(4.0_double, 5.0 * sqrt(backgroundCountsMode))
        end if 
    else
        maxHeight = backgroundCountsMode + max(4.0_double, 5.0 * sqrt(backgroundCountsMode))
    end if
    do i = il, iu
        do j = jl, ju
            if(image(i, j) .gt. maxHeight) then
                imageOkay = .false.
                return
            end if

        end do
    end do

    imageOkay = .true.

end function imageOkay

!***********************************************************
!
!
!
!
!**********************************************************
subroutine computeTheBackground(image, nsigma, background, sum, nPixels)
    real(single) :: image(:, :)
    real(single), intent(in) :: nSigma
    real(single) :: background(:)
    real(double) :: sum(:)
    integer, intent(Out) :: nPixels
    integer :: il, iu, jl, ju, i, j, n1

    real(single) :: sigma
    real(single), pointer :: pixel(:)=> null()
 
    call setBackgroundNsigma( nSigma ) ! In mssllib, omskybackground_mod

    il = lbound(image, 1)
    iu = ubound(image, 1)
    jl = lbound(image, 2)
    ju = ubound(image, 2)
    

    n1 = (iu - il + 1) * (ju - jl + 1)
    pixel => allocateSingle(1, n1, "pixel")

    sum = 0.0_double
    nPixels = 0 ! The number of pixels

    do i = il, iu
        do j = jl, ju
            if (.not. isNullValue(image(i, j) ) ) then
                nPixels = nPixels + 1
                pixel(nPixels) = image(i, j)
                sum = sum + dble(image(i, j))
            end if 
        end do
    end do

    if(nPixels >= 50) then

        !***********************************************************************
        !
        ! Loop through all the different background methods
        !
        !***********************************************************************
        do i = 1, nMethods

            call setBackgroundMethod(int(i, int8))  ! In mssllib, omskybackground_mod

            background(i) = BackgroundCountsPerPixel(nPixels, pixel, &
                      nSigma, sigma)

        end do
    else
        nPixels = -1
    end if   

    deallocate(pixel)
    nullify(pixel)

end subroutine computeTheBackground


!*******************************************************************
!
!
!*****************************************************************
subroutine drawGraphs1( image, title )  
    real(single), pointer :: image(:, :)
    character(*), intent(in) :: title
    integer :: nx, ny, i, j, nPixels   
    real(single), pointer :: pixels(:) => null()
    real(single), pointer, dimension(:) :: background(:) => null()
    real(single) :: nSigma, mean, std, sigma
    real(single) :: binwidth1, yMin, yMax, y1, y2, dy, dx, snValue
    real(double) :: binWidth, xmin, xmax
    integer :: noBins 
    character(200) :: text


    write(errstr, '(a)' ) "Computing background value (counts per pixel) for the whole image"

    call message(AppMsg, Verbose)


    nx = size(image, 1)
    ny = size(image, 2)

    nPixels = nx * ny
    pixels => allocateSingle(1, nPixels, "pixels")
    nx = nx - 1
    ny = ny - 1
    nPixels = 0
    do i = 0, nx
        do j = 0, ny
            if (.not. isNullValue(image(i, j) ) ) then
                nPixels = nPixels + 1
                pixels(nPixels) = image(i, j)
            end if
        end do
    end do

    if(nPixels .lt. 50) then
        deallocate( pixels )
        return
    end if


    call ComputeAverage( pixels, nPixels, mean, std)
   

    snValue = Sn(pixels, nPixels)

    !*************************************************************************
    !
    !
    !**********************************************************************   
    

    do i = 6, 10
        noBins = 50


        xmin = Percentile(nPixels, pixels,  0.1 - real(i - 1) * 0.01)
        xmax = Percentile(nPixels, pixels,  0.9 + real(i - 1) * 0.01)
  
   
    
      

        if( abs(xMin) < 1e-6) then
            if( (xmax - xmin) >= 3.0) then
                exit   
            end if

        else
        
            if( ((xmax - xmin) / abs(xmin)) >= 2.0) then
                exit
            end if
        end if

    end do



 
    background => allocateSingle(1, int(nMethods, int32), "background")
    sigma = 2.0


    do i = 1, nMethods


        call setBackgroundMethod(int(i, int8))  ! In mssllib, omskybackground_mod

        background(i) = BackgroundCountsPerPixel(nPixels, pixels, nSigma, sigma)

 
        write(errstr, '(a, a, a, f0.2, a, f0.2)' ) "Background method = ", trim(getBackgroundText(i)), &
                    " computed background=", background(i), "+/-", sigma
        call message(AppMsg, Verbose)


    end do

    if (.not. plotGraph ) then

        deallocate( pixels, background )
        return
    end if

   
    !*****************************************************
    !
    !
    !*****************************************************

  
    binWidth = getBinWidth(xmin, xmax, 1.0_double)


    xmin = xmin - binwidth / 2.0_double
    xmax = xmax + binwidth / 2.0_double



    binWidth1 = real(binWidth)


    call Histogram(nPixels, pixels, binWidth1, &
                 "Pixel counts", trim(title), 0.1, 0.9, 0.1, 0.7, &
                  xmin=real(xmin), xmax=real(xmax), colour=getRed() )



    !*****************************************************
    !
    ! Subroutine DrawPoissonDistribution
    !
    !*****************************************************
    call pgqwin(xmin, xmax, ymin, ymax)
    


    y1  = 0.0
    dy = (ymax - ymin) / dble(nMethods)
    dx = (xmax - xmin) / dble(20)
    
    write(text, '(a, f0.2, a, f0.2)')  "Robust scale estimator sn=", snValue, " standard deviation=", std
    call pgslw(3)
    call pgsci(1)
    call pgsch(0.9)
    call pgmtxt('T', -1.0, 0.4, 0.0, trim(text))              
    
    do i = 1, nMethods

        write(text, '(a, a, f0.2, a, f0.2)')  trim(getBackgroundText(i)), "=", background(i), "+/-", sigma
!        call pgsci( i )
        call pgsch(0.9)
        call pgslw(3)

        y1 = real(i - 1) * dy
        y2 = real(i) * dy
        
        call pgmtxt('T', -real(i + 1), 0.4, 0.0, trim(text))              
!        call pgmove(background(i), y1)
!        call pgdraw(background(i), y2)
!        y1 = (y1 + y2) / 2.0
!        call pgmove(background(i) - dx, y1)
!        call pgdraw(background(i) + dx, y1)

    end do

    call pgsci( getblack ())



    call DrawPoissonDistribution(nPixels, binWidth, dble(background(1) ), getMediumBlue() )


!    call DrawPoissonCurve(nPixels, binWidth, dble(background(1)), mediumBlue)

    



    deallocate( pixels, background )

    call pgpage()

end subroutine drawGraphs1


!*******************************************************************
!
!
!*****************************************************************
subroutine drawGraphs2( background, nSize, xLabel, title )  
    real(single), pointer :: background(:, :)
    integer, intent(in) :: nSize
    character(*), intent(in) :: xLabel, title
  
    integer(int32) :: nData(10)
    real(single) :: mean, std, binWidth1
    real(double) :: binWidth, xmin, xmax, p1, p2
    character(100) :: caption(10)
    integer :: i, j, noBins

    nData = nSize
  

    caption = " "
   
  
    noBins = 50

  

    do j = 6, 10 
        xmin = huge(xmin)
        xmax = -xmin

  


        do i = 1, nMethods 

            p1 = Percentile(nSize, background(:, i), 0.1 - real(i - 1) * 0.01)
            p2 = Percentile(nSize, background(:, i), 0.9 + real(j - 1) * 0.01)

            xMin = min(xMin, p1)
            xMax = max(xMax, p2)
     
        end do


        if( abs(xMin) < 1e-6 ) then
            if( (xmax - xmin) >= 3.0) then
                exit   
            end if

        else
        
            if( ((xmax - xmin) / abs(xmin)) >= 2.0) then
                exit
            end if
        end if

    end do
 

    binWidth = getBinWidth(xmin, xmax, 0.1_double)



    do i = 1, nMethods
        call ComputeAverage( background(:, i), nSize, mean, std)
        p1 = Percentile(nSize, background(:, i), 0.05)
        p2 = Percentile(nSize, background(:, i), 0.95)




        write(caption(i), '(a, a, f0.2, a, f0.2, a, f0.2, a, f0.2)')  trim(getBackgroundText(i)), "=", &
                mean, "+/-", std, ", 0.05 percentile=", p1, ", 0.95=", p2 
             
    end do

 
    xmin = xmin - binwidth / 2.0_double
    xmax = xmax + binwidth / 2.0_double



    binWidth1 = real(binWidth)

    call Histogram61(int(nMethods, int32), ndata(1 : nMethods), &
                 background(1 : nSize, 1 : nMethods), binWidth1, &
                 xLabel, title, caption(1 : nMethods), 0.1, 0.9, 0.1, 0.7, &
                 real(xmin), real(xmax))



    call pgpage()

end subroutine drawGraphs2
!***************************************************************
!
!
!****************************************************************
real(double) function getBinWidth(xmin, xmax, minBinSize)
    real(double), intent(inout) :: xMin, xMax
    real(double), intent(in), optional :: minBinSize
    real(double) :: binWidth1, constant
    integer :: noBins

    integer :: i, j
    real(double) :: binwidth(8) = (/0.1_double, 0.2_double, 0.25_double, &
                                   0.5_double, 1.0_double, 2.0_double,  &
                                   5.0_double, 10.0_double/)




     xmin = floor( xmin * 1000.0 ) / 1000.0 

     xmax = ceiling( xmax * 1000.0 ) / 1000.0




jLoop  : do i = 0, 10
            constant = 10.0_double ** i 
        do j = 1, 8       
             
             binWidth1 = constant * binWidth(j)
             if(present(minBinSize)) then
                 if(binWidth1 < minBinSize) then
                     cycle

                 end if
             end if 

             noBins = int( ( (xmax - xmin) / binwidth1) + 0.01)
             if(noBins <= 20) then
          
                 exit jLoop    

            end if
        end do     
    end do jLoop

    getBinWidth = binWidth1

    i = floor(xMin / binWidth1)
    

    xMin = binWidth1 * dble(i)

    i = ceiling(xMax / binWidth1)


   
    xMax = binWidth1 * dble(i)

    
    noBins = int( ( (xmax - xmin) / binwidth1) + 0.01)

    if(noBins .lt. 6) then

        i = 6 - noBins
        xmax = xmax + dble(i) * binWidth1 

    end if   




end function getBinWidth
!***************************************************************
!
!
!
!**************************************************************
subroutine createTheBackgroundImage( image, imageFile, backgroundImageFile )

    real(single), pointer :: image(:, :)
    character(*), intent(in) :: imageFile, backgroundImageFile
    integer(int16) :: nx, ny, width
   
    real(single), pointer, dimension(:, :) :: backgroundImage => null(), imageP => null()

   
    real(single) :: sigma, background1
    type(DataSetT) :: inDataSetH, outDataSetH
    type(ArrayT) :: outArrayH                                         ! Output array handle.

    nx = ubound(image, 1) 
    ny = ubound(image, 2)

    backgroundImage  => allocate2dSingle(0_int16, nx, 0_int16, ny, "backgroundImage ")
 

  

    nx = size(image, 1) 
    ny = size(image, 2)


    width = intParameter("boxwidth") * 2 / binx
    backgroundImage  = image
    

    call createBackgroundImage5( backgroundImage , width)

    background1 = ComputeBackgroundMode2( backgroundImage , nx, ny, 3.0, sigma)


    write(errstr, '(a, f0.2, a, f0.2)' ) &
          "Global background of background image=", background1, "+/-", sigma
    call message(AppMsg, Verbose)


   


    !******************************************************************
    !
    !
    !******************************************************************

    inDataSetH = dataSet( imageFile, Read)

    outDataSetH = dataSet( backgroundImageFile , Create)
    call copyAllAttributes(attributable(inDataSetH), attributable(outDataSetH))

    call release( inDataSetH )

  
  
   


    outArrayH = addArray( outDataSetH, "PRIMARY", real32, shape(image))

    imageP => real32Array2Data( outArrayH )

    imageP = backgroundImage

 
 
    call release( outArrayH ) 
    call release( outDataSetH )




    deallocate(backgroundImage)
    


end subroutine createTheBackgroundImage

end module ombackground_mod





