! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
module robustscale_mod

use omallocation_mod
use omnumerics_mod
implicit none
contains
!******************************************************************************
!
!   This file contains Fortran code for two new estimators of scale
!   proposed in Rousseeuw and Croux (1993). The estimators are
!
!       Sn = cn * 1.1926 * LOMED_{i} HIMED_{i} |x_i-x_j|
!
!   and
!
!       Qn = dn * 2.2219 * {|x_i-x_j|; i<j}_(k)
!
!   The values of the constants cn, qn, and k are already included
!   in the source code. The algorithms for Sn and Qn are very fast.
!   They only consume O(nlogn) time and use O(n) storage space.
!
!   The estimators Sn and Qn can be used as alternatives to the
!   median absolute deviation (MAD) because, like the MAD,
!
!      - they have an explicit formula (hence they are unique);
!      - their breakdown point is still 50%;
!      - their influence functions are still bounded;
!      - their bias curves are also very low;
!      - the estimates can be computed fast;
!
!   whereas they have the additional advantages of:
!
!      - a higher efficiency (58% for Sn and 82% for Qn,
!        compared to 37% for the MAD);
!      - being suitable for symmetric as well as asymmetric
!        distributions, because Sn and Qn are not based on the
!        absolute deviations from a central value.
!
!   Choosing between Sn and Qn depends on the situation. On the
!   one hand, Qn is more efficient than Sn. On the other hand,
!   Sn is more robust in the sense of a smaller bias and a lower
!   gross-error sensitivity. As a default choice, we recommend
!   using the estimator Sn.
!
!   More information can be found in the paper:
!
!   Rousseeuw, P.J. and Croux, C. (1993), "Alternatives to the
!      Median Absolute Deviation," Journal of the American
!      Statistical Association, Vol. 88, 1273-1283.
!
!   This software may be used and copied freely, provided
!   reference is made to the abovementioned paper.
!
!   For questions, problems or comments contact:
!
!              Peter Rousseeuw (rousse@wins.uia.ac.be)
!              Christophe Croux (croux@wins.uia.ac.be)
!              Department of Mathematics and Computing
!              Universitaire Instelling Antwerpen
!              Universiteitsplein 1
!              B-2610 Wilrijk (Antwerp)
!              Belgium
!
!
!   The sources attached below are:
!
!      Sn.for    Fortran code for the computation of Sn
!      Qn.for    Fortran code for the computation of Qn
!
!   If you would like Pascal versions of the algorithms,
!   please contact us directly.
!
!
!#####################################################################
!######################  file Sn.for :  ##############################
!#####################################################################
!
!   This file contains a Fortran function for a new robust estimator
!   of scale denoted as Sn, proposed in Rousseeuw and Croux (1993).
!   The estimator has a high breakdown point and a bounded influence
!   function. The algorithm given here is very fast (running in
!   O(nlogn) time) and needs only O(n) storage space.
!
!   Rousseeuw, P.J. and Croux, C. (1993), "Alternatives to the
!      Median Absolute Deviation," Journal of the American
!      Statistical Association, Vol. 88, 1273-1283.
!
!   A Fortran function for the estimator Qn, described in the same
!   paper, is attached below. For both estimators, implementations
!   in the Pascal language can be obtained from the authors.
!
!   This software may be used and copied freely, provided
!   reference is made to the abovementioned paper.
!
!   For questions, problems or comments contact:
!
!              Peter Rousseeuw (rousse@wins.uia.ac.be)
!              Christophe Croux (croux@wins.uia.ac.be)
!              Department of Mathematics and Computing
!              Universitaire Instelling Antwerpen
!              Universiteitsplein 1
!              B-2610 Wilrijk (Antwerp)
!              Belgium
!
!--------------------------------------------------------------------
!
!   Efficient algorithm for the scale estimator:
!
!       Sn = cn * 1.1926 * LOMED_{i} HIMED_{i} |x_i-x_j|
!
!   which can equivalently be written as
!
!       Sn = cn * 1.1926 * LOMED_{i} LOMED_{j<>i} |x_i-x_j|
!
!   Parameters of the function Sn :
!       x  : real array containing the observations
!       n  : number of observations (n>=2)
!
!   The function Sn uses the procedures:
!       sort(x,n,y) : sorts an array x of length n, and stores the
!                     result in an array y (of size at least n)
!       pull(a,n,k) : finds the k-th order statistic of an
!                     array a of length n
!
!   The function Sn also creates an auxiliary array a2
!       (of size at least n) in which it stores the values
!       LOMED_{j<>i} |x_i-x_j|   for i=1,...,n
!
!*****************************************************************************
real(single) function Sn(x, n)
    real(single) ::  x(:)
    integer, intent(in) :: n

    real(single), pointer, dimension(:) :: y => null(), a2 => null()
    integer :: rightA, rightB, tryA, tryB, diff, Amin, Amax, even, half
    real(single) :: medA, medB
    integer :: i, length, na, nb, lefta, leftb, cn

    y => allocateSingle(1, n, "y")
    a2 => allocateSingle(1, n, "a2")

!write(*, *) "sn line 131 n=", n  

    y = x(1 : n)
    call heapsort(y, n)

    a2(1) = y(n / 2 + 1) - y(1)

    do 10 i = 2,(n + 1) / 2
        nA = i - 1
        nB = n - i
        diff= nB - nA
        leftA = 1
        leftB = 1
        rightA = nB
        rightB = nB
        Amin = diff / 2 + 1
        Amax = diff / 2 + nA
15      continue
        if (leftA .lt. rightA) then
            length = rightA - leftA + 1
            even = 1 - mod(length, 2)
            half = (length - 1) / 2
            tryA = leftA + half
            tryB = leftB + half
            if (tryA .lt. Amin) then
                rightB = tryB
                leftA = tryA + even
            else
                if (tryA.gt.Amax) then
                    rightA = tryA
                    leftB = tryB + even
                else
                    medA = y(i) - y(i - tryA + Amin - 1)
                    medB = y(tryB + i) - y(i)
                    if (medA .ge. medB) then
                        rightA = tryA
                        leftB = tryB + even
                    else
                        rightB = tryB
                        leftA = tryA + even
                    endif
                endif
            endif
            go to 15
        endif
        if (leftA .gt. Amax) then
            a2(i) = y(leftB + i) - y(i)
        else
            medA = y(i) - y(i - leftA + Amin - 1)
            medB = y(leftB + i) - y(i)
            a2(i) = min(medA, medB)
        endif
10      continue
      do 20 i=(n+1)/2+1,n-1
          nA=n-i
          nB=i-1
          diff=nB-nA
          leftA=1
          leftB=1
          rightA=nB
          rightB=nB
          Amin=diff/2+1
          Amax=diff/2+nA
25        continue
          if (leftA.lt.rightA) then
              length=rightA-leftA+1
              even=1-mod(length,2)
              half=(length-1)/2
              tryA=leftA+half
              tryB=leftB+half
              if (tryA.lt.Amin) then
                  rightB=tryB
                  leftA=tryA+even
              else
                  if (tryA.gt.Amax) then
                      rightA=tryA
                      leftB=tryB+even
                  else
                      medA=y(i+tryA-Amin+1)-y(i)
                      medB=y(i)-y(i-tryB)
                      if (medA.ge.medB) then
                          rightA=tryA
                          leftB=tryB+even
                      else
                          rightB=tryB
                          leftA=tryA+even
                      endif
                  endif
              endif
          go to 25
          endif
          if (leftA.gt.Amax) then
              a2(i)=y(i)-y(i-leftB)
          else
              medA=y(i+leftA-Amin+1)-y(i)
              medB=y(i)-y(i-leftB)
              a2(i)=min(medA,medB)
          endif
20    continue
      a2(n)=y(n)-y((n+1)/2)
      cn=1
      if (n.le.9) then
          if (n.eq.2) cn=0.743
          if (n.eq.3) cn=1.851
          if (n.eq.4) cn=0.954
          if (n.eq.5) cn=1.351
          if (n.eq.6) cn=0.993
          if (n.eq.7) cn=1.198
          if (n.eq.8) cn=1.005
          if (n.eq.9) cn=1.131
      else
          if (mod(n,2).eq.1) cn=n/(n-0.9)
      endif
      Sn=cn*1.1926*pull(a2,n,(n+1)/2)
!write(*, *) "leaving sn"
    deallocate(y, a2)      
end function sn

!**************************************************************************
!
!   This file contains a Fortran function for a new robust estimator
!   of scale denoted as Qn, proposed in Rousseeuw and Croux (1993).
!   The estimator has a high breakdown point and a smooth and bounded
!   influence function. The algorithm given here is very fast (running
!   in O(nlogn) time) and needs only O(n) storage space.
!
!   Rousseeuw, P.J. and Croux, C. (1993), "Alternatives to the
!      Median Absolute Deviation," Journal of the American
!      Statistical Association, Vol. 88, 1273-1283.
!
!   A Fortran function for the estimator Sn, described in the same
!   paper, is attached above. For both estimators, implementations
!   in the Pascal language can be obtained from the authors.
!
!   This software may be used and copied freely, provided
!   reference is made to the abovementioned paper.
!
!   For questions, problems or comments contact:
!
!              Peter Rousseeuw (rousse@wins.uia.ac.be)
!              Christophe Croux (croux@wins.uia.ac.be)
!              Department of Mathematics and Computing
!              Universitaire Instelling Antwerpen
!              Universiteitsplein 1
!              B-2610 Wilrijk (Antwerp)
!              Belgium
!
!--------------------------------------------------------------------
!
!   Efficient algorithm for the scale estimator:
!
!       Qn = dn * 2.2219 * {|x_i-x_j|; i<j}_(k)
!
!   Parameters of the function Qn :
!       x  : real array containing the observations
!       n  : number of observations (n >=2)
!
!   The function Qn uses the procedures:
!      whimed(a,iw,n): finds the weighted high median of an array
!                      a of length n, using the array iw (also of
!                      length n) with positive integer weights.
!      sort(x,n,y) : sorts an array x of length n, and stores the
!                    result in an array y (of size at least n)
!      pull(a,n,k) : finds the k-th order statistic of an
!                    array a of length n
!
!*******************************************************************
 
real(single) function Qn(x, n)
    integer, intent(in) :: n     
    real(single) ::  x(n)
    real(single), pointer, dimension(:) :: y => null(), work => null()

    integer, pointer, dimension(:) ::  left=>null(), right=>null(), &
           weight=>null(), Q=>null(), P=>null()

    integer :: h, k, knew, jhelp, nL, nR, sumQ, sumP, i, j, jj
    real(single) :: dn, trial
    logical found
    

    y => allocateSingle(1, n, "y")
    y = x(1 : n)
    work => allocateSingle(1, 2 * n, "work")

    left => allocateInt32(1, n, "left")
    right => allocateInt32(1, n, "right")
    weight => allocateInt32(1, n, "weight")
    Q => allocateInt32(1, n, "Q")
    P => allocateInt32(1, n, "left")


!    write(*, *) "qn line 331- n=", n

    h = n / 2 + 1
    k = h * (h - 1) / 2
    call heapsort(y, n)

    do i = 1, n
        left(i) = n - i + 2
        right(i) = n
    end do 

    jhelp = n * (n + 1) / 2
    knew = k + jhelp
    nL = jhelp
    nR = n * n
    found = .false.

!write(*, *) "Line 347"


200   continue
  
!write(*, *) "line 352"
    if ( (nR - nL .gt. n) .and. (.not.found) ) then
        j = 1

!write(*, *) "Line 354"

        do i = 2, n
            if (left(i) .le. right(i)) then
                weight(j) = right(i) - left(i) + 1
                jhelp = left(i) + weight(j) / 2
                work(j) = y(i) - y(n + 1 - jhelp)
                j = j + 1
            endif
        end do
        trial = whimed(work, weight, j - 1)
        j = 0

!write(*, *) "Line 368"


          do i = n, 1, -1
45            continue
!          write(*, *) i, ") j=", j, " n=", n, " trial=", trial
              if ((j .lt. n - 1) .and.((y(i) - y(n - j)) .lt. trial)) then
                  j = j + 1
                  goto 45
              endif
              P(i) = j

          end do

!write(*, *) "Line 391"

          j = n + 1
          do i = 1,  n
55            continue
              if ((y(i) - y(n - j + 2)) .gt. trial) then
                  j = j - 1
                  goto 55
              endif
              Q(i) = j
          end do
          sumP=0
          sumQ=0

!write(*, *) "Line 391"



          do i = 1, n
              sumP = sumP + P(i)
              sumQ = sumQ + Q(i) - 1
          end do
          if (knew .le. sumP) then
              do i =1, n
                  right(i) = P(i)
              end do
              nR = sumP
          else
              if (knew.gt.sumQ) then
                  do i=1,n
                      left(i)=Q(i)
                  end do
                  nL=sumQ
              else
                  Qn=trial
                  found=.true.
              endif
          endif
      goto 200
      endif

!write(*, *) "Line 418"

    if (.not. found) then
        j = 1
        do i = 2, n
!           write(*, *) i, ") j=", j, " left=", left(i), " right=", right(i)
            if (left(i) .le. right(i)) then
                do jj = left(i), right(i)
                    work(j) = y(i) - y(n - jj + 1)
                    j = j + 1
                end do
            endif
        end do
!write(*, *) "Line 441"

        Qn = pull(work, j - 1, knew - nL)
!write(*, *) "Line 444"

    endif

    if (n .le. 9) then
        if (n .eq. 2) dn = 0.399
        if (n .eq. 3) dn = 0.994
        if (n .eq. 4) dn = 0.512
        if (n .eq. 5) dn = 0.844
        if (n .eq. 6) dn = 0.611
        if (n .eq. 7) dn = 0.857
        if (n .eq. 8) dn = 0.669
        if (n .eq. 9) dn = 0.872
    else
        if (mod(n, 2) .eq. 1) dn = n / (n + 1.4)
        if (mod(n, 2) .eq. 0) dn = n / (n + 3.8)
    endif
    Qn = dn * 2.2219 * Qn

!write(*, *) "leaving qn"
    deallocate(y, work, left, right, weight, q, p)
   
end function qn



!************************************************************************
!
!
!  Algorithm to compute the weighted high median in O(n) time.
!
!  The whimed is defined as the smallest a(j) such that the sum
!  of the weights of all a(i) <= a(j) is strictly greater than
!  half of the total weight.
!
!  Parameters of this function:
!        a: real array containing the observations
!        n: number of observations
!       iw: array of integer weights of the observations.
!
!  This function uses the function pull.
!
!  The size of acand, iwcand must be at least n.
!
!***************************************************************************** 
real(single) function whimed(a, iw, n)
    integer, intent(in) :: n
    real(single) :: a(n)
    integer :: iw(n)
    real(single), pointer, dimension(:) :: acand => null(), iwcand => null() 

  
    integer :: wtotal, wrest, wleft, wmid, wright
    integer :: i, nn, kCand
    real(single) :: trial

    acand => allocateSingle(1, n, "acand")
    iwcand => allocateSingle(1, n, "iwcand")

    nn = n
    wtotal = 0
    do i= 1, nn
        wtotal = wtotal + iw(i)
    end do
    wrest = 0
100   continue
    trial = pull(a, nn, nn / 2 +1)
    wleft = 0
    wmid = 0
    wright = 0
    do i = 1, nn
        if (a(i) .lt. trial) then
            wleft=wleft+iw(i)
        else
            if (a(i).gt.trial) then
                wright=wright+iw(i)
            else
                wmid=wmid+iw(i)
            endif
        endif
    end do
    if ((2 * wrest + 2 * wleft) .gt. wtotal) then
        kcand = 0
        do i = 1, nn
            if (a(i) .lt. trial) then
                kcand = kcand + 1
                acand(kcand) = a(i)
                iwcand(kcand) = iw(i)
            end if
        end do
        nn = kcand
    else
        if ((2 * wrest + 2 * wleft + 2 * wmid) .gt. wtotal) then
            whimed = trial
            deallocate(acand)
            deallocate(iwcand)
            return
        else
            kcand = 0
            do i = 1, nn
                if(a(i) .gt. trial) then
                    kcand = kcand + 1
                    acand(kcand) = a(i)
                    iwcand(kcand) = iw(i)
                endif
            end do
            nn = kcand
            wrest = wrest + wleft + wmid
        end if
    endif
    do i = 1, nn
        a(i) = acand(i)
        iw(i) = iwcand(i)
    end do

    go to 100

end function whimed


!***************************************************************
!
! Finds the kth order statistic of an array a of length n<=1000
!
!**************************************************************

real(single) function pull(a, n, k)
    integer, intent(in) :: n, k
      real(single) :: a(n)

      real(single), pointer :: b(:)
      real(single) :: ax, buffer
      integer :: j, l, lr, jnc

      b => allocateSingle(1, n, "b")
      b = a(1 : n)

      l=1
      lr=n
20    if (l.ge.lr) goto 90
      ax=b(k)
      jnc=l
      j=lr
30    if(jnc.gt.j)goto 80
40    if (b(jnc).ge.ax)goto 50
      jnc=jnc+1
      goto 40
50    if (b(j).le.ax)goto 60
      j=j-1
      goto 50
60    if(jnc.gt.j)goto 70
      buffer=b(jnc)
      b(jnc)=b(j)
      b(j)=buffer
      jnc=jnc+1
      j=j-1
70    goto 30
80    if(j.lt.k)l=jnc
      if(k.lt.jnc)lr=j
      goto 20
90    pull=b(k)
     
      deallocate(b) 
    
end function pull

end module robustscale_mod
